import abc
from itertools import permutations

from guarddog.analyzer.metadata.detector import Detector


class TyposquatDetector(Detector):
    MESSAGE_TEMPLATE = "This package closely ressembles the following package names, and might be a typosquatting " \
                       "attempt: %s"

    def __init__(self) -> None:
        self.popular_packages = self._get_top_packages()  # Find top PyPI packages
        super().__init__(
            name="typosquatting",
            description="Identify packages that are named closely to an highly popular package"
        )

    @abc.abstractmethod
    def _get_top_packages(self) -> list:
        pass

    def _is_distance_one_Levenshtein(self, name1, name2) -> bool:
        """
        Returns True if two names have a Levenshtein distance of one

        Args:
            name1 (str): first name
            name2 (str): second name

        Returns:
            bool: True if within distance one
        """

        if abs(len(name1) - len(name2)) > 1:
            return False

        # Addition to name2
        if len(name1) > len(name2):
            for i in range(len(name1)):
                if name1[:i] + name1[i + 1:] == name2:
                    return True

        # Addition to name1
        elif len(name2) > len(name1):
            for i in range(len(name2)):
                if name2[:i] + name2[i + 1:] == name1:
                    return True

        # Edit character
        else:
            for i in range(len(name1)):
                if name1[:i] + name1[i + 1:] == name2[:i] + name2[i + 1:]:
                    return True

        return False

    def _is_swapped_typo(self, name1, name2) -> bool:
        """
        Returns true is two names are adjacent swaps of each other

        Args:
            name1 (str): first name
            name2 (str): second name

        Returns:
            bool: True if adjacent swaps
        """

        if len(name1) == len(name2):
            for i in range(len(name1) - 1):
                swapped_name1 = name1[:i] + name1[i + 1] + name1[i] + name1[i + 2:]
                if swapped_name1 == name2:
                    return True

        return False

    def _generate_permutations(self, package_name) -> list[str]:
        """
        Generates all permutations of hyphenated terms of a package

        Args:
            package_name (str): name of package

        Returns:
            list[str]: permutations of package_name
        """

        if "-" not in package_name:
            return []

        components = package_name.split("-")
        hyphen_permutations = ["-".join(p) for p in permutations(components)]

        return hyphen_permutations

    def _is_length_one_edit_away(self, package1, package2) -> bool:
        """
        Returns True if two packages are within a distance one typo edit
        (either within a Levenshtein distance of one or an adjacent swap edit)

        Args:
            package1 (str): first package name
            package2 (str): second package name

        Returns:
            bool: True
        """

        return self._is_distance_one_Levenshtein(package1, package2) or self._is_swapped_typo(package1, package2)

    def _get_confused_forms(self, package_name) -> list:
        """
        Gets confused terms for python packages
        Confused terms are:
            - py to python swaps (or vice versa)
            - the removal of py/python terms

        Args:
            package_name (str): name of the package

        Returns:
            list: list of confused terms
        """

        confused_forms = []

        terms = package_name.split("-")

        # Detect swaps like python-package -> py-package
        for i in range(len(terms)):
            confused_term = None

            if "python" in terms[i]:
                confused_term = terms[i].replace("python", "py")
            elif "py" in terms[i]:
                confused_term = terms[i].replace("py", "python")
            else:
                continue

            # Get form when replacing or removing py/python term
            replaced_form = terms[:i] + [confused_term] + terms[i + 1:]
            removed_form = terms[:i] + terms[i + 1:]

            for form in (replaced_form, removed_form):
                confused_forms.append("-".join(form))

        return confused_forms

    def get_typosquatted_package(self, package_name) -> list[str]:
        """
        Gets all legitimate packages that a given name
        is possibly typosquatting from

        Checks for Levenshtein distance, permutations, and confused terms
        against the top 5000 most downloaded PyPI packages

        Args:
            package_name (str): name of package

        Returns:
            list[str]: names of packages that <package_name> could be
            typosquatting from
        """

        if package_name in self.popular_packages:
            return []

        # Go through popular packages and find length one edit typosquats
        typosquatted = set()
        for popular_package in self.popular_packages:
            if self._is_length_one_edit_away(package_name, popular_package):
                typosquatted.add(popular_package)

            alternate_popular_names = self._get_confused_forms(popular_package)
            swapped_popular_names = self._generate_permutations(popular_package)

            for name in alternate_popular_names + swapped_popular_names:
                if self._is_length_one_edit_away(package_name, name):
                    typosquatted.add(popular_package)

        return list(typosquatted)



""" PyPI Package Malware Scanner

CLI command that scans a PyPI package version for user-specified malware flags.
Includes rules based on package registry metadata and source code analysis.
"""
import logging
import os
import sys
from typing import cast, Optional

import click
from prettytable import PrettyTable
from termcolor import colored

from guarddog.analyzer.analyzer import SEMGREP_RULE_NAMES
from guarddog.analyzer.metadata import get_metadata_detectors
from guarddog.analyzer.sourcecode import SOURCECODE_RULES
from guarddog.ecosystems import ECOSYSTEM
from guarddog.reporters.sarif import report_verify_sarif
from guarddog.scanners import get_scanner
from guarddog.scanners.scanner import PackageScanner

ALL_RULES = \
    set(get_metadata_detectors(ECOSYSTEM.NPM).keys()) \
    | set(get_metadata_detectors(ECOSYSTEM.PYPI).keys()) | SEMGREP_RULE_NAMES
NPM_RULES = set(get_metadata_detectors(ECOSYSTEM.NPM).keys()) | SEMGREP_RULE_NAMES
PYPI_RULES = set(get_metadata_detectors(ECOSYSTEM.PYPI).keys()) | SEMGREP_RULE_NAMES
EXIT_CODE_ISSUES_FOUND = 1

AVAILABLE_LOG_LEVELS = {
    logging.DEBUG,
    logging.INFO,
    logging.WARN,
    logging.ERROR
}
AVAILABLE_LOG_LEVELS_NAMES = list(map(lambda level: logging.getLevelName(level), AVAILABLE_LOG_LEVELS))

log = logging.getLogger('guarddog')


def common_options(fn):
    fn = click.option("--exit-non-zero-on-finding", default=False, is_flag=True,
                      help="Exit with a non-zero status code if at least one issue is identified")(fn)
    fn = click.argument("target")(fn)
    return fn


def legacy_rules_options(fn):
    fn = click.option("-r", "--rules", multiple=True, type=click.Choice(ALL_RULES, case_sensitive=False))(fn)
    fn = click.option("-x", "--exclude-rules", multiple=True, type=click.Choice(ALL_RULES, case_sensitive=False))(fn)
    return fn


def npm_options(fn):
    fn = click.option("-r", "--rules", multiple=True, type=click.Choice(NPM_RULES, case_sensitive=False))(fn)
    fn = click.option("-x", "--exclude-rules", multiple=True, type=click.Choice(NPM_RULES, case_sensitive=False))(fn)
    return fn


def pypi_options(fn):
    fn = click.option("-r", "--rules", multiple=True, type=click.Choice(PYPI_RULES, case_sensitive=False))(fn)
    fn = click.option("-x", "--exclude-rules", multiple=True, type=click.Choice(PYPI_RULES, case_sensitive=False))(fn)
    return fn


def verify_options(fn):
    fn = click.option("--output-format", default=None, type=click.Choice(["json", "sarif"], case_sensitive=False))(fn)
    return fn


def scan_options(fn):
    fn = click.option("--output-format", default=None, type=click.Choice(["json"], case_sensitive=False))(fn)
    fn = click.option("-v", "--version", default=None, help="Specify a version to scan")(fn)
    return fn


def logging_options(fn):
    fn = click.option("--log-level", default="INFO",
                      type=click.Choice(AVAILABLE_LOG_LEVELS_NAMES, case_sensitive=False))(fn)
    return fn


@click.group
@logging_options
def cli(log_level):
    """
    GuardDog cli tool to detect malware in package ecosystems

    Supports PyPI and npm

    Example: guarddog pypi scan semantic-version

    Use --help for the detail of all commands and subcommands
    """
    logger = logging.getLogger('guarddog')
    logger.setLevel(logging.getLevelName(log_level))
    stdoutHandler = logging.StreamHandler(stream=sys.stdout)
    stdoutHandler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    logger.addHandler(stdoutHandler)
    pass


def _get_rule_param(rules, exclude_rules, ecosystem):
    rule_param = None
    if len(rules) > 0:
        rule_param = rules

    if len(exclude_rules) > 0:
        all_rules = SEMGREP_RULE_NAMES
        if ecosystem == ECOSYSTEM.NPM:
            all_rules |= set(get_metadata_detectors(ECOSYSTEM.NPM).keys())
        elif ecosystem == ECOSYSTEM.PYPI:
            all_rules |= set(get_metadata_detectors(ECOSYSTEM.PYPI).keys())

        rule_param = all_rules - set(exclude_rules)

        if len(rules) > 0:
            print("--rules and --exclude-rules cannot be used together")
            exit(1)

    return rule_param


def _verify(path, rules, exclude_rules, output_format, exit_non_zero_on_finding, ecosystem):
    """Verify a requirements.txt file

    Args:
        path (str): path to requirements.txt file
    """
    return_value = None
    rule_param = _get_rule_param(rules, exclude_rules, ecosystem)
    scanner = get_scanner(ecosystem, True)
    if scanner is None:
        sys.stderr.write(f"Command verify is not supported for ecosystem {ecosystem}")
        exit(1)

    def display_result(result: dict) -> None:
        identifier = result['dependency'] if result['version'] is None \
            else f"{result['dependency']} version {result['version']}"
        if output_format is None:
            print_scan_results(result.get('result'), identifier)

        if len(result.get('errors', [])) > 0:
            print_errors(result.get('error'), identifier)

    results = scanner.scan_local(path, rule_param, display_result)
    if output_format == "json":
        import json as js
        return_value = js.dumps(results)

    if output_format == "sarif":
        sarif_rules = PYPI_RULES if ecosystem == ECOSYSTEM.PYPI else NPM_RULES
        return_value = report_verify_sarif(path, list(sarif_rules), results, ecosystem)

    if output_format is not None:
        print(return_value)

    if exit_non_zero_on_finding:
        exit_with_status_code(results)

    return return_value  # this is mostly for testing


def is_local_target(identifier: str) -> bool:
    """
    @param identifier:  The name/path of the package as passed to "guarddog ecosystem scan"
    @return:            Whether the identifier should be consider a local path
    """
    if identifier.startswith("/") or identifier.startswith("./") or identifier.startswith("../"):
        return True

    if identifier == ".":
        return True

    # If this looks like an archive, consider it as a local target if the target exists on the local filesystem
    if identifier.endswith(".tar.gz") or identifier.endswith(".zip") or identifier.endswith(".whl"):
        return os.path.exists(identifier)

    return False


def _scan(identifier, version, rules, exclude_rules, output_format, exit_non_zero_on_finding, ecosystem: ECOSYSTEM):
    """Scan a package

    Args:
        identifier (str): name or path to the package
        version (str): version of the package (ex. 1.0.0), defaults to most recent
        rules (list[str]): specific rules to run, defaults to all
    """

    rule_param = _get_rule_param(rules, exclude_rules, ecosystem)
    scanner = cast(Optional[PackageScanner], get_scanner(ecosystem, False))
    if scanner is None:
        sys.stderr.write(f"Command scan is not supported for ecosystem {ecosystem}")
        exit(1)
    results = []
    if is_local_target(identifier):
        log.debug(f"Considering that '{identifier}' is a local target, scanning filesystem")
        if os.path.isdir(identifier):
            log.debug(f"Considering that '{identifier}' as a local directory")
            for package in os.listdir(identifier):
                results.append({'package': package} | scanner.scan_local(f"{identifier}/{package}", rule_param))
        else:
            results.append({'package': identifier} | scanner.scan_local(identifier, rule_param))
    else:
        log.debug(f"Considering that '{identifier}' is a remote target")
        try:
            results.append({'package': identifier} | scanner.scan_remote(identifier, version, rule_param))
        except Exception as e:
            sys.stderr.write("\n")
            sys.stderr.write(str(e))
            sys.exit()

    if output_format == "json":
        import json as js
        if len(results) == 1:
            # return only a json like {}
            print(js.dumps(results[0]))
        else:
            # Return a list of result like [{},{}]
            print(js.dumps(results))
    else:
        for result in results:
            print_scan_results(result, result['package'])

    if exit_non_zero_on_finding:
        exit_with_status_code(results)


def _list_rules(ecosystem):
    table = PrettyTable()
    table.align = "l"
    table.field_names = ["Rule type", "Rule name", "Description"]

    for rule in SOURCECODE_RULES[ecosystem]:
        table.add_row(["Source code", rule['id'], rule.get('metadata', {}).get('description')])

    metadata_rules = get_metadata_detectors(ecosystem)
    for ruleName in metadata_rules:
        rule = metadata_rules[ruleName]
        table.add_row(["Package metadata", rule.get_name(), rule.get_description()])

    print(table)


@cli.group
def npm(**kwargs):
    """ Scan a npm package or verify a npm project
    """
    pass


@cli.group
def pypi(**kwargs):
    """ Scan a PyPI package or verify a PyPI project
    """
    pass


@npm.command("scan")
@common_options
@scan_options
@npm_options
def scan_npm(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """ Scan a given npm package
    """
    return _scan(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.NPM)


@npm.command("verify")
@common_options
@verify_options
@npm_options
def verify_npm(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """ Verify a given npm project
    """
    return _verify(target, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.NPM)


@pypi.command("scan")
@common_options
@scan_options
@pypi_options
def scan_pypi(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """ Scan a given PyPI package
    """
    return _scan(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.PYPI)


@pypi.command("verify")
@common_options
@verify_options
@pypi_options
def verify_pypi(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """ Verify a given Pypi project
    """
    return _verify(target, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.PYPI)


@pypi.command("list-rules")
def list_rules_pypi():
    """ Print available rules for PyPI
    """
    return _list_rules(ECOSYSTEM.PYPI)


@npm.command("list-rules")
def list_rules_npm():
    """ Print available rules for npm
    """
    return _list_rules(ECOSYSTEM.NPM)


@cli.command("verify", deprecated=True)
@common_options
@verify_options
@legacy_rules_options
def verify(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    return _verify(target, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.PYPI)


@cli.command("scan", deprecated=True)
@common_options
@scan_options
@legacy_rules_options
def scan(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    return _scan(target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding, ECOSYSTEM.PYPI)


# Pretty prints scan results for the console
def print_scan_results(results, identifier):
    num_issues = results.get('issues')
    errors = results.get('errors', [])

    if num_issues == 0:
        print("Found " + colored('0 potentially malicious indicators', 'green',
                                 attrs=['bold']) + " scanning " + colored(identifier, None, attrs=['bold']))
        print()
    else:
        print("Found " + colored(str(num_issues) + ' potentially malicious indicators', 'red',
                                 attrs=['bold']) + " in " + colored(identifier, None, attrs=['bold']))
        print()

        findings = results.get('results', [])
        for finding in findings:
            description = findings[finding]
            if type(description) == str:  # package metadata
                print(colored(finding, None, attrs=['bold']) + ': ' + description)
                print()
            elif type(description) == list:  # semgrep rule result:
                source_code_findings = description
                print(colored(finding, None,
                              attrs=['bold']) + ': found ' + str(len(source_code_findings)) + ' source code matches')
                for finding in source_code_findings:
                    print('  * ' + finding['message']
                          + ' at ' + finding['location'] + '\n    ' + format_code_line_for_output(finding['code']))
                print()

    if len(errors) > 0:
        print_errors(errors, identifier)
        print('\n')


def print_errors(errors, identifier):
    print(colored("Some rules failed to run while scanning " + identifier + ":", "yellow"))
    print()
    for rule in errors:
        print(f'* {rule}: {errors[rule]}')
    print()


def format_code_line_for_output(code):
    return '    ' + colored(code.strip().replace('\n', '\n    ').replace('\t', '  '), None, 'on_red', attrs=['bold'])


# Given the results, exit with the appropriate status code
def exit_with_status_code(results):
    for result in results:
        num_issues = result.get('issues', 0)
        if num_issues > 0:
            exit(EXIT_CODE_ISSUES_FOUND)



from guarddog.analyzer.metadata.pypi.empty_information import PypiEmptyInfoDetector
from guarddog.analyzer.metadata.pypi.potentially_compromised_email_domain import \
    PypiPotentiallyCompromisedEmailDomainDetector
from guarddog.analyzer.metadata.pypi.release_zero import PypiReleaseZeroDetector
from guarddog.analyzer.metadata.pypi.repository_integrity_mismatch import PypiIntegrityMismatchDetector
from guarddog.analyzer.metadata.pypi.single_python_file import PypiSinglePythonFileDetector
from guarddog.analyzer.metadata.pypi.typosquatting import PypiTyposquatDetector

PYPI_METADATA_RULES = {}

classes = [
    PypiEmptyInfoDetector,
    PypiReleaseZeroDetector,
    PypiTyposquatDetector,
    PypiPotentiallyCompromisedEmailDomainDetector,
    PypiIntegrityMismatchDetector,
    PypiSinglePythonFileDetector

]

for detectorClass in classes:
    detectorInstance = detectorClass()  # type: ignore
    PYPI_METADATA_RULES[detectorInstance.get_name()] = detectorInstance



from typing import Optional, Any
from pathlib import Path
import json

from guarddog.analyzer.metadata.detector import Detector

# List of fields where mismatch between package.json and NPM can carry malicious information
# (field, expected type)
MANIFEST_FIELDS_CHECKLIST = {
    "dependencies": dict,
    "devDependencies": dict,
    "scripts": dict,
    "main": str,
    "repository": dict,
    "bugs": dict,
    "homepage": str
}


class NPMMetadataMismatch(Detector):
    def __init__(self):
        super().__init__(
            name="npm_metadata_mismatch",
            description="Identify packages which have mismatches between the npm pacakge manifest and the package info"
        )

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, Optional[str]]:
        # Get the latest version if not specified
        if not version:
            version = package_info["dist-tags"]["latest"]

        # Load package.json manifest
        if path is None:
            raise ValueError("path is needed to run heuristic " + self.get_name())
        package_json = Path(path) / "package" / "package.json"
        package_manifest: dict[str, Any] = json.loads(package_json.read_text())

        # Get NPM manifest for version
        version_info = package_info["versions"][version]

        diff: dict[str, Diff] = {
            field: difference_at_key(version_info, package_manifest, field, field_type)
            for field, field_type in MANIFEST_FIELDS_CHECKLIST.items()
        }
        number_different = sum(len(v) for k, v in diff.items())
        diff_description = describe_diff(diff) if number_different != 0 else "No differences found"
        return number_different != 0, diff_description


# PerItemDiff is (key, left_value, right_value)
PerItemDiff = tuple[str, Optional[str], Optional[str]]
Diff = list[PerItemDiff]


def diff_at_key_dict(version_at_key: dict[str, Optional[str]], manifest_at_key: dict[str, Optional[str]]) -> Diff:
    return [
        (key, version_at_key.get(key), manifest_at_key.get(key))
        for key in set(version_at_key.keys()).union(set(manifest_at_key.keys()))
        if version_at_key.get(key) != manifest_at_key.get(key)
    ]


def difference_at_key(version_info: dict[str, Any], package_manifest: dict[str, Any], key: str, key_type) -> Diff:
    version_at_key = version_info.get(key, key_type())
    manifest_at_key = package_manifest.get(key, key_type())
    if not (isinstance(version_at_key, key_type) and isinstance(manifest_at_key, key_type)):
        return [(f"Expected type {str(key_type)}", f"{type(version_at_key)}", f"{type(manifest_at_key)}")]
    elif key_type == dict:
        return diff_at_key_dict(version_at_key, manifest_at_key)
    else:
        # If it is not a dict do a direct comparison of the value at the key, currently the only other type is strings
        return [(f"{key}", version_at_key, manifest_at_key)] if version_at_key != manifest_at_key else []


def describe_diff(diff: dict[str, Diff]) -> str:
    """
    Creates a string of the form
    Difference between manifest and package.json found:
    dependencies:
        key: Manifest("v4.0.0"), package.json("v3.0.1")
    scripts:
        key: Manifest("a"), package.json("b")
    main:
        Manifest:
            index.js
        package.json
            malicious.js
    ...
    """
    description = "Difference between manifest and package.json found: \n"
    for k, differences in diff.items():
        if differences:
            field_description = f"{k}: \n"
            if MANIFEST_FIELDS_CHECKLIST[k] == dict:
                for d in differences:
                    field_description += f"  {d[0]}: Manifest(\"{d[1]}\"), package.json(\"{d[2]}\") \n"
            else:
                manifest_str = "  Manifest:\n"
                package_str = "  package.json:\n"
                for d in differences:
                    manifest_str += f"    {d[1]}\n"
                    package_str += f"    {d[2]}\n"
                field_description = field_description + manifest_str + package_str
            description += field_description
    return description



""" Empty Information Detector

Detects when a package has its latest release version to 0.0.0
"""
import logging
from typing import Optional

from guarddog.analyzer.metadata.release_zero import ReleaseZeroDetector

log = logging.getLogger("guarddog")


class PypiReleaseZeroDetector(ReleaseZeroDetector):

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, str]:
        log.debug(f"Running zero version heuristic on PyPI package {name} version {version}")
        return (package_info["info"]["version"] in ["0.0.0", "0.0"],
                ReleaseZeroDetector.MESSAGE_TEMPLATE % package_info["info"]["version"])



import json
import logging

import requests
from semantic_version import NpmSpec, Version  # type:ignore

from guarddog.scanners.npm_package_scanner import NPMPackageScanner
from guarddog.scanners.scanner import ProjectScanner

log = logging.getLogger("guarddog")


def find_all_versions(package_name: str, semver_range: str) -> set[str]:
    url = f"https://registry.npmjs.org/{package_name}"
    log.debug(f"Retrieving npm package metadata from {url}")
    response = requests.get(url)
    if response.status_code != 200:
        log.debug(f"No version available, status code {response.status_code}")
        return set()

    data = response.json()
    versions = list(data["versions"].keys())
    log.debug(f"Retrieved versions {', '.join(versions)}")
    result = set()
    try:
        npm_spec = NpmSpec(semver_range)
    except ValueError:  # not a semver range, let's keep it raw
        result.add(semver_range)
        return result
    for v in versions:
        if Version(v) in npm_spec:
            result.add(v)
    return result


class NPMRequirementsScanner(ProjectScanner):
    """
    Scans all packages in the package.json file of a project

    Attributes:
        package_scanner (PackageScanner): Scanner for individual packages
    """

    def __init__(self) -> None:
        super().__init__(NPMPackageScanner())

    def parse_requirements(self, raw_requirements: str) -> dict:
        package = json.loads(raw_requirements)
        dependencies = package["dependencies"] if "dependencies" in package else {}
        dev_dependencies = package["devDependencies"] if "devDependencies" in package else {}

        merged = {}  # type: dict[str, set[str]]
        for package, selector in list(dependencies.items()) + list(dev_dependencies.items()):
            if package not in merged:
                merged[package] = set()
            merged[package].add(selector)

        results = {}
        for package, all_selectors in merged.items():
            versions = set()  # type: set[str]
            for selector in all_selectors:
                versions = versions.union(find_all_versions(package, selector))
            if len(versions) > 0:
                results[package] = versions
        return results



import json
import os
from typing import Optional

from guarddog.analyzer.metadata.typosquatting import TyposquatDetector


class NPMTyposquatDetector(TyposquatDetector):
    """Detector for typosquatting attacks. Detects if a package name is a typosquat of one of the top 5000 packages.
    Checks for distance one Levenshtein, one-off character swaps, permutations
    around hyphens, and substrings.

    Attributes:
        popular_packages (list): list of top 5000 downloaded packages from npm
    """

    def _get_top_packages(self) -> list:
        top_packages_filename = "top_npm_packages.json"
        resources_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))
        with open(os.path.join(resources_dir, top_packages_filename)) as file:
            top_packages_data = json.load(file)

        return list(map(lambda x: x["project"], top_packages_data))

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, Optional[str]]:
        """
        Uses a package's information from PyPI's JSON API to determine the
        package is attempting a typosquatting attack

        Args:
            package_info (dict): dictionary representation of PyPI's JSON
                output

        Returns:
            list[str]: names of packages that <package_name> could be
            typosquatting from
            @param **kwargs:
        """

        similar_package_names = self.get_typosquatted_package(package_info["name"])
        if len(similar_package_names) > 0:
            return True, TyposquatDetector.MESSAGE_TEMPLATE % ", ".join(similar_package_names)
        return False, None



import json
import logging
import os
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Iterable, List, Optional

from guarddog.analyzer.metadata import get_metadata_detectors
from guarddog.ecosystems import ECOSYSTEM


def get_rules(file_extension, path):
    return set(rule.replace(file_extension, "") for rule in os.listdir(path) if rule.endswith(file_extension))


SEMGREP_RULES_PATH = os.path.join(os.path.dirname(__file__), "sourcecode")
SEMGREP_RULE_NAMES = get_rules(".yml", SEMGREP_RULES_PATH)

log = logging.getLogger("guarddog")


class Analyzer:
    """
    Analyzes a local directory for threats found by source code or metadata rules

    Attributes:
        sourcecode_rules_path (str): path to source code rules
        ecosystem (str): name of the current ecosystem
        metadata_ruleset (list): list of metadata rule names
        sourcecode_ruleset (list): list of source code rule names

        exclude (list): list of directories to exclude from source code search

        metadata_detectors(list): list of metadata detectors
    """

    def __init__(self, ecosystem=ECOSYSTEM.PYPI) -> None:
        self.sourcecode_rules_path = os.path.join(os.path.dirname(__file__), "sourcecode")

        self.ecosystem = ecosystem

        # Rules and associated detectors
        self.metadata_detectors = get_metadata_detectors(ecosystem)

        self.metadata_ruleset = self.metadata_detectors.keys()
        self.sourcecode_ruleset = SEMGREP_RULE_NAMES

        # Define paths to exclude from sourcecode analysis
        self.exclude = [
            "helm",
            ".idea",
            "venv",
            "test",
            "tests",
            ".env",
            "dist",
            "build",
            "semgrep",
            "migrations",
            ".github",
            ".semgrep_logs",
        ]

    def analyze(self, path, info=None, rules=None, name: Optional[str] = None, version: Optional[str] = None) -> dict:
        """
        Analyzes a package in the given path

        Args:
            path (str): path to package
            info (dict, optional): Any package information to analyze metadata. Defaults to None.
            rules (set, optional): Set of rules to analyze. Defaults to all rules.

        Raises:
            Exception: "{rule} is not a valid rule."

        Returns:
            dict[str]: map from each rule and their corresponding output
        """

        metadata_results = None
        sourcecode_results = None

        # populate results, errors, and number of issues
        metadata_rules = None
        sourcecode_rules = None
        if rules is not None:
            # Only run specific rules
            sourcecode_rules = set()
            metadata_rules = set()

            for rule in rules:
                if rule in self.sourcecode_ruleset:
                    log.debug(f"Using source code rule {rule}")
                    sourcecode_rules.add(rule)
                elif rule in self.metadata_ruleset:
                    log.debug(f"Using metadata rule {rule}")
                    metadata_rules.add(rule)
                else:
                    raise Exception(f"{rule} is not a valid rule.")

        log.debug(f"Running metadata rules against package '{name}'")
        metadata_results = self.analyze_metadata(path, info, metadata_rules, name, version)

        log.debug(f"Running source code rules against directory '{path}'")
        sourcecode_results = self.analyze_sourcecode(path, sourcecode_rules)

        # Concatenate dictionaries together
        issues = metadata_results["issues"] + sourcecode_results["issues"]
        results = metadata_results["results"] | sourcecode_results["results"]
        errors = metadata_results["errors"] | sourcecode_results["errors"]

        return {"issues": issues, "errors": errors, "results": results, "path": path}

    def analyze_metadata(self, path: str, info, rules=None, name: Optional[str] = None,
                         version: Optional[str] = None) -> dict:
        """
        Analyzes the metadata of a given package

        Args:
            path (str): path to package
            info (dict): package information given by PyPI Json API
            rules (set, optional): Set of metadata rules to analyze. Defaults to all rules.

        Returns:
            dict[str]: map from each metadata rule and their corresponding output
        """

        all_rules = rules if rules is not None else self.metadata_ruleset
        results = {}
        errors = {}
        issues = 0

        for rule in all_rules:
            try:
                log.debug(f"Running rule {rule} against package '{name}'")
                rule_matches, message = self.metadata_detectors[rule].detect(info, path, name, version)
                if rule_matches:
                    issues += 1
                    results[rule] = message
            except Exception as e:
                errors[rule] = f"failed to run rule {rule}: {str(e)}"

        return {"results": results, "errors": errors, "issues": issues}

    def analyze_sourcecode(self, path, rules=None) -> dict:
        """
        Analyzes the source code of a given package

        Args:
            path (str): path to directory of package
            rules (set, optional): Set of source code rules to analyze. Defaults to all rules.

        Returns:
            dict[str]: map from each source code rule and their corresponding output
        """
        targetpath = Path(path)
        all_rules = rules if rules is not None else self.sourcecode_ruleset
        results = {rule: {} for rule in all_rules}  # type: dict
        errors = {}
        issues = 0

        rules_path: List[str]
        if rules is None:
            log.debug(f"No rules specified using full rules directory {self.sourcecode_rules_path}")
            rules_path = [self.sourcecode_rules_path]
        else:
            rules_path = list(map(
                lambda rule_name: os.path.join(self.sourcecode_rules_path, f"{rule_name}.yml"),
                rules
            ))

        if len(rules_path) == 0:
            log.debug("No source code rules to run")
            return {"results": {}, "errors": {}, "issues": 0}

        try:
            log.debug(f"Running source code rules against {path}")
            response = self._invoke_semgrep(target=path, rules=rules_path)
            rule_results = self._format_semgrep_response(response, targetpath=targetpath)
            issues += sum(len(res) for res in rule_results.values())

            results = results | rule_results
        except Exception as e:
            errors["rules-all"] = f"failed to run rule: {str(e)}"

        return {"results": results, "errors": errors, "issues": issues}

    def _invoke_semgrep(self, target: str, rules: Iterable[str]):
        try:
            cmd = ["semgrep"]
            for rule in rules:
                cmd.extend(["--config", rule])

            for excluded in self.exclude:
                cmd.append(f"--exclude='{excluded}'")
            cmd.append("--no-git-ignore")
            cmd.append("--json")
            cmd.append("--quiet")
            cmd.append(target)
            log.debug(f"Invoking semgrep with command line: {' '.join(cmd)}")
            result = subprocess.run(cmd, capture_output=True, check=True, encoding="utf-8")
            return json.loads(str(result.stdout))
        except FileNotFoundError:
            raise Exception("unable to find semgrep binary")
        except subprocess.CalledProcessError as e:
            error_message = f"""
An error occurred when running Semgrep.

command: {" ".join(e.cmd)}
status code: {e.returncode}
output: {e.output}
"""
            raise Exception(error_message)
        except json.JSONDecodeError as e:
            raise Exception("unable to parse semgrep JSON output: " + str(e))

    def _format_semgrep_response(self, response, rule=None, targetpath=None):
        """
        Formats the response from Semgrep

        Args:
            response (dict): response from Semgrep
            rule (str, optional): name of rule to format. Defaults to all rules.
            targetpath (str, optional): root directory of scan. Defaults to None.
                Paths in formatted resonse will be rooted from targetpath.

        Returns:
            dict: formatted response in the form...

            {
                ...
                <rule-name>: [
                    {
                        <path-to-code:line-num>: <dangerous-code>
                        ...
                    },
                    ...
                ],
                ...
            }
        """

        results = defaultdict(list)

        for result in response["results"]:
            rule_name = rule or result["check_id"].split(".")[-1]
            code_snippet = result["extra"]["lines"]
            line = result["start"]["line"]

            file_path = os.path.abspath(result["path"])
            if targetpath:
                file_path = os.path.relpath(file_path, targetpath)

            location = file_path + ":" + str(line)
            code = self.trim_code_snippet(code_snippet)

            results[rule_name].append({
                'location': location,
                'code': code,
                'message': result["extra"]["message"]
            })

        return results

    # Makes sure the matching code to be displayed isn't too long
    def trim_code_snippet(self, code):
        THRESHOLD = 250
        if len(code) > THRESHOLD:
            return code[: THRESHOLD - 10] + '...' + code[len(code) - 10:]
        else:
            return code



from abc import abstractmethod
from typing import Optional

from guarddog.analyzer.metadata.detector import Detector


class IntegrityMismatch(Detector):
    """This package contains files that have been tampered with between the source repository and the package CDN"""
    RULE_NAME = "repository_integrity_mismatch"

    def __init__(self):
        super().__init__(
            name="repository_integrity_mismatch",
            description="Identify packages with a linked GitHub repository where the package has extra unexpected files"
        )

    @abstractmethod
    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, str]:
        pass



from typing import Optional

from .npm_package_scanner import NPMPackageScanner
from .npm_project_scanner import NPMRequirementsScanner
from .pypi_package_scanner import PypiPackageScanner
from .pypi_project_scanner import PypiRequirementsScanner
from .scanner import Scanner
from ..ecosystems import ECOSYSTEM


def get_scanner(ecosystem: ECOSYSTEM, project: bool) -> Optional[Scanner]:
    match (ecosystem, project):
        case (ECOSYSTEM.PYPI, False):
            return PypiPackageScanner()
        case (ECOSYSTEM.PYPI, True):
            return PypiRequirementsScanner()
        case (ECOSYSTEM.NPM, False):
            return NPMPackageScanner()
        case (ECOSYSTEM.NPM, True):
            return NPMRequirementsScanner()
    return None



from enum import Enum


class ECOSYSTEM(Enum):
    PYPI = "pypi"
    NPM = "npm"


def get_friendly_name(ecosystem: ECOSYSTEM) -> str:
    match ecosystem:
        case ECOSYSTEM.PYPI:
            return "PyPI"
        case ECOSYSTEM.NPM:
            return "npm"
        case _:
            return ecosystem.value



import logging
import os
import zipfile

import tarsafe  # type:ignore

log = logging.getLogger("guarddog")


def safe_extract(source_archive: str, target_directory: str) -> None:
    """
    safe_extract safely extracts archives to a target directory.

    This function does not clean up the original archive, and does not create the target directory if it does not exist.

    @param source_archive:      The archive to extract
    @param target_directory:    The directory where to extract the archive to
    @raise ValueError           If the archive type is unsupported
    """
    log.debug(f"Extracting archive {source_archive} to directory {target_directory}")
    if source_archive.endswith('.tar.gz') or source_archive.endswith('.tgz'):
        tarsafe.open(source_archive).extractall(target_directory)
    elif source_archive.endswith('.zip') or source_archive.endswith('.whl'):
        with zipfile.ZipFile(source_archive, 'r') as zip:
            for file in zip.namelist():
                # Note: zip.extract cleans up any malicious file name such as directory traversal attempts
                # This is not the case of zipfile.extractall
                zip.extract(file, path=os.path.join(target_directory, file))
    else:
        raise ValueError("unsupported archive extension: " + target_directory)



import os
import typing

from guarddog.analyzer.analyzer import Analyzer
from guarddog.ecosystems import ECOSYSTEM
from guarddog.scanners.scanner import PackageScanner
from guarddog.utils.package_info import get_package_info


class PypiPackageScanner(PackageScanner):
    def __init__(self) -> None:
        super().__init__(Analyzer(ECOSYSTEM.PYPI))

    def download_and_get_package_info(self, directory: str, package_name: str, version=None) -> typing.Tuple[dict, str]:
        extract_dir = self.download_package(package_name, directory, version)
        return get_package_info(package_name), extract_dir

    def download_package(self, package_name, directory, version=None) -> str:
        """Downloads the PyPI distribution for a given package and version

        Args:
            package_name (str): name of the package
            directory (str): directory to download package to
            version (str): version of the package

        Raises:
            Exception: "Received status code: " + <not 200> + " from PyPI"
            Exception: "Version " + version + " for package " + package_name + " doesn't exist."
            Exception: "Compressed file for package does not exist."
            Exception: "Error retrieving package: " + <error message>
        Returns:
            Path where the package was extracted
        """

        data = get_package_info(package_name)
        releases = data["releases"]

        if version is None:
            version = data["info"]["version"]

        if version in releases:
            files = releases[version]

            url = None
            file_extension = None

            for file in files:
                # Store url to compressed package and appropriate file extension
                if file["filename"].endswith(".tar.gz"):
                    url = file["url"]
                    file_extension = ".tar.gz"

                if file["filename"].endswith(".egg") or file["filename"].endswith(".whl") \
                        or file["filename"].endswith(".zip"):
                    url = file["url"]
                    file_extension = ".zip"

            if url and file_extension:
                # Path to compressed package
                zippath = os.path.join(directory, package_name + file_extension)
                unzippedpath = zippath.removesuffix(file_extension)

                self.download_compressed(url, zippath, unzippedpath)
                return unzippedpath
            else:
                raise Exception(f"Compressed file for {package_name} does not exist on PyPI.")
        else:
            raise Exception("Version " + version + " for package " + package_name + " doesn't exist.")



import logging
import re
import sys

import pkg_resources
import requests

from guarddog.scanners.pypi_package_scanner import PypiPackageScanner
from guarddog.scanners.scanner import ProjectScanner

log = logging.getLogger("guarddog")


class PypiRequirementsScanner(ProjectScanner):
    """
    Scans all packages in the requirements.txt file of a project

    Attributes:
        package_scanner (PackageScanner): Scanner for individual packages
    """

    def __init__(self) -> None:
        super().__init__(PypiPackageScanner())

    def _sanitize_requirements(self, requirements: list[str]) -> list[str]:
        """
        Filters out non-requirement specifications from a requirements specification

        Args:
            requirements (str): PEP440 styled dependency specification text

        Returns:
            list[str]: sanitized lines containing only version specifications
        """

        sanitized_lines = []

        for line in requirements:
            is_requirement = re.match(r"\w", line)
            if is_requirement:
                if "\\" in line:
                    line = line.replace("\\", "")

                stripped_line = line.strip()
                if len(stripped_line) > 0:
                    sanitized_lines.append(stripped_line)

        return sanitized_lines

    # FIXME: type return value properly to dict[str, set[str]]
    def parse_requirements(self, raw_requirements: str) -> dict:
        """
        Parses requirements.txt specification and finds all valid
        versions of each dependency

        Args:
            requirements (List[str]): contents of requirements.txt file

        Returns:
            dict: mapping of dependencies to valid versions

            ex.
            {
                ....
                <dependency-name>: [0.0.1, 0.0.2, ...],
                ...
            }
        """
        requirements = raw_requirements.splitlines()

        def versions(package_name):
            url = "https://pypi.org/pypi/%s/json" % (package_name,)
            log.debug(f"Retrieving PyPI package metadata information from {url}")
            data = requests.get(url).json()
            versions = sorted(data["releases"].keys(), reverse=True)
            return versions

        sanitized_requirements = self._sanitize_requirements(requirements)

        dependencies = {}

        def safe_parse_requirements(req):
            parsed = pkg_resources.parse_requirements(req)
            while True:
                try:
                    yield next(parsed)
                except StopIteration:
                    break
                except Exception as e:
                    sys.stderr.write(f"Error when parsing requirements, received error {str(e)}. This entry will be "
                                     "ignored.\n")
                    yield None

        try:
            for requirement in safe_parse_requirements(sanitized_requirements):
                if requirement is None:
                    continue
                valid_versions = None
                project_exists_on_pypi = True
                for spec in requirement.specs:
                    qualifier, version = spec

                    try:
                        available_versions = versions(requirement.project_name)  # type: list[str]
                    except Exception:
                        sys.stderr.write(f"Package {requirement.project_name} not on PyPI\n")
                        project_exists_on_pypi = False
                        continue

                    used_versions = None

                    match qualifier:
                        case ">":
                            used_versions = {v for v in available_versions if v > version}
                        case "<":
                            used_versions = {v for v in available_versions if v < version}
                        case ">=":
                            used_versions = {v for v in available_versions if v >= version}
                        case "<=":
                            used_versions = {v for v in available_versions if v <= version}
                        case "==":
                            matches = [re.search(version, candidate) for candidate in available_versions]
                            filtered_matches = list(filter(None, matches))
                            str_matches = [v.string for v in filtered_matches]
                            used_versions = set(str_matches)
                        case "~=":
                            prefix = "".join(version.split(".")[:-1])
                            for available_version in available_versions:  # sorted decreasing
                                if available_version >= version and available_version.startswith(prefix):
                                    used_versions = set(available_version)
                                    break
                        case _:
                            sys.stderr.write(f"Unknown qualifier: {qualifier}")
                            continue

                    if valid_versions is None:
                        valid_versions = used_versions
                    else:
                        valid_versions = valid_versions & used_versions

                if project_exists_on_pypi:
                    dependencies[requirement.project_name] = valid_versions
        except Exception as e:
            sys.stderr.write(f"Received error {str(e)}")

        return dependencies



""" Direct URL Dependency Detector

Detects if a package depends on direct URL dependencies
"""
from typing import Optional
import re

from guarddog.analyzer.metadata.detector import Detector

from urllib.parse import urlparse


github_project_pattern = re.compile(r"^([\w\-\.]+)/([\w\-\.]+)")


class NPMDirectURLDependencyDetector(Detector):
    """This heuristic detects packages with direct URL dependencies.
    Dependencies fetched this way are not immutable and can be used to inject untrusted code
    or reduce the likelihood of a reproducible install."""

    def __init__(self):
        super().__init__(
            name="direct_url_dependency",
            description="Identify packages with direct URL dependencies. \
Dependencies fetched this way are not immutable and can be used to \
inject untrusted code or reduce the likelihood of a reproducible install.",
        )

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        findings = []

        for dep_name, dep_version in (
            package_info.get("versions", {})
            .get(version, {})
            .get("dependencies", {})
            .items()
        ):
            # According to npm documentation, HTTP(s) and Git are accepted URL schemes when specifying dependencies:
            # https://docs.npmjs.com/cli/v10/configuring-npm/package-json#urls-as-dependencies
            # https://docs.npmjs.com/cli/v10/configuring-npm/package-json#git-urls-as-dependencies
            if urlparse(dep_version).scheme in [
                "http",
                "https",
                "git",
                "git+ssh",
                "git+http",
                "git+https",
                "git+file",
            ]:
                findings.append(
                    f"Dependency {dep_name} refers to a direct Git or HTTP URL {dep_version}."
                )
            # According to npm documentation, Github repositories are accepted when specifying dependencies:
            # https://docs.npmjs.com/cli/v10/configuring-npm/package-json#github-urls
            elif github_project_pattern.match(dep_version):
                findings.append(
                    f"Dependency {dep_name} refers to a direct Github repository {dep_version}."
                )

        return len(findings) != 0, "\n".join(findings)



""" Empty Information Detector

Detects if a package contains an empty description
"""
import logging
from typing import Optional

from guarddog.analyzer.metadata.empty_information import EmptyInfoDetector

MESSAGE = "This package has an empty description on PyPi"

log = logging.getLogger("guarddog")


class PypiEmptyInfoDetector(EmptyInfoDetector):
    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, str]:
        log.debug(f"Running PyPI empty description heuristic on package {name} version {version}")
        return len(package_info["info"]["description"].strip()) == 0, EmptyInfoDetector.MESSAGE_TEMPLATE % "PyPI"



import os
import pathlib

import yaml
from yaml.loader import SafeLoader

from guarddog.ecosystems import ECOSYSTEM

current_dir = pathlib.Path(__file__).parent.resolve()
rule_file_names = list(
    filter(
        lambda x: x.endswith('yml'),
        os.listdir(current_dir)
    )
)

SOURCECODE_RULES = {
    ECOSYSTEM.PYPI: list(),
    ECOSYSTEM.NPM: list()
}  # type: dict[ECOSYSTEM, list[dict]]

for file_name in rule_file_names:
    with open(os.path.join(current_dir, file_name), "r") as fd:
        data = yaml.load(fd, Loader=SafeLoader)
        for rule in data["rules"]:
            for lang in rule["languages"]:
                match lang:
                    case "python":
                        SOURCECODE_RULES[ECOSYSTEM.PYPI].append(rule)
                    case "javascript" | "typescript" | "json":
                        SOURCECODE_RULES[ECOSYSTEM.NPM].append(rule)



""" Empty Information Detector

Detects if a package contains an empty description
"""
import os.path
from typing import Optional

from guarddog.analyzer.metadata.empty_information import EmptyInfoDetector

MESSAGE = "This package has an empty description on PyPi"


class NPMEmptyInfoDetector(EmptyInfoDetector):

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, str]:
        if path is None:
            raise TypeError("path must be a string")
        package_path = os.path.join(path, "package")
        content = map(
            lambda x: x.lower(),
            os.listdir(package_path)
        )
        return "readme.md" not in content, EmptyInfoDetector.MESSAGE_TEMPLATE % "npm"



import json
import logging
import os
from datetime import datetime, timedelta
from typing import Optional

import requests
import packaging.utils

from guarddog.analyzer.metadata.typosquatting import TyposquatDetector


log = logging.getLogger("guarddog")


TOP_PACKAGES_CACHE_LOCATION = os.environ.get('GUARDDOG_TOP_PACKAGES_CACHE_LOCATION')


class PypiTyposquatDetector(TyposquatDetector):
    """
    Detector for typosquatting attacks. Detects if a package name is a typosquat of one of the top 1000 packages.
    Checks for distance one Levenshtein, one-off character swaps, permutations
    around hyphens, and substrings.

    Attributes:
        popular_packages (list): list of top 5000 downloaded packages from PyPI
    """

    def _get_top_packages(self) -> list:
        """
        Gets the package information of the top 5000 most downloaded PyPI packages

        Returns:
            list: list of package data in the format:
                [
                    ...
                    {
                        download_count: ...
                        project: <package-name>
                    }
                    ...
                ]
        """

        popular_packages_url = "https://hugovk.github.io/top-pypi-packages/top-pypi-packages-30-days.min.json"

        top_packages_filename = "top_pypi_packages.json"
        resources_dir = TOP_PACKAGES_CACHE_LOCATION
        if resources_dir is None:
            resources_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))

        top_packages_path = os.path.join(resources_dir, top_packages_filename)

        top_packages_information = None

        if top_packages_filename in os.listdir(resources_dir):
            update_time = datetime.fromtimestamp(os.path.getmtime(top_packages_path))

            if datetime.now() - update_time <= timedelta(days=30):
                with open(top_packages_path, "r") as top_packages_file:
                    top_packages_information = json.load(top_packages_file)["rows"]

        if top_packages_information is None:
            response = requests.get(popular_packages_url).json()
            with open(top_packages_path, "w+") as f:
                json.dump(response, f, ensure_ascii=False, indent=4)

            top_packages_information = response["rows"]

        def get_safe_name(package):
            return packaging.utils.canonicalize_name(package["project"])

        return list(map(get_safe_name, top_packages_information))

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, Optional[str]]:
        """
        Uses a package's information from PyPI's JSON API to determine the
        package is attempting a typosquatting attack

        Args:
            package_info (dict): dictionary representation of PyPI's JSON
                output

        Returns:
            list[str]: names of packages that <package_name> could be
            typosquatting from
            @param **kwargs:
        """
        log.debug(f"Running typosquatting heuristic on PyPI package {name}")
        normalized_name = packaging.utils.canonicalize_name(package_info["info"]["name"])
        similar_package_names = self.get_typosquatted_package(normalized_name)
        if len(similar_package_names) > 0:
            return True, TyposquatDetector.MESSAGE_TEMPLATE % ", ".join(similar_package_names)
        return False, None



""" Empty Information Detector

Detects when a package has its latest release version to 0.0.0
"""
from typing import Optional

from guarddog.analyzer.metadata.release_zero import ReleaseZeroDetector


class NPMReleaseZeroDetector(ReleaseZeroDetector):

    def detect(self, package_info, path: Optional[str] = None, name: Optional[str] = None,
               version: Optional[str] = None) -> tuple[bool, str]:
        return package_info["dist-tags"]["latest"] in ["0.0.0", "0.0", "0"], \
            ReleaseZeroDetector.MESSAGE_TEMPLATE % package_info["dist-tags"]["latest"]



import logging

import requests

log = logging.getLogger("guarddog")


def get_package_info(name: str) -> dict:
    """Gets metadata and other information about package

    Args:
        name (str): name of the package

    Raises:
        Exception: "Received status code: " + str(response.status_code) + " from PyPI"
        Exception: "Error retrieving package: " + data["message"]

    Returns:
        json: package attributes and values
    """

    url = "https://pypi.org/pypi/%s/json" % (name,)
    log.debug(f"Retrieving PyPI package metadata from {url}")
    response = requests.get(url)

    # Check if package file exists
    if response.status_code != 200:
        raise Exception("Received status code: " + str(response.status_code) + " from PyPI")

    data = response.json()

    # Check for error in retrieving package
    if "message" in data:
        raise Exception("Error retrieving package: " + data["message"])

    return data



"""Default execution entry point if running the package via python -m."""

import sys

import guarddog.cli


def cli():
    """Run guarddog from script entry point"""
    return guarddog.cli.cli()


if __name__ == "__main__":
    sys.exit(cli())



""" Compromised Email Detector

Detects if a maintainer's email domain might have been compromised.
"""

from datetime import datetime
from typing import Optional

from dateutil import parser

from guarddog.analyzer.metadata.potentially_compromised_email_domain import PotentiallyCompromisedEmailDomainDetector


class NPMPotentiallyCompromisedEmailDomainDetector(PotentiallyCompromisedEmailDomainDetector):
    def __init__(self):
        super().__init__("npm")

    def get_email_addresses(self, package_info: dict) -> list[str]:
        if package_info.get("maintainers"):
            return list(map(lambda x: x["email"], package_info["maintainers"]))
        else:
            return []

    def get_project_latest_release_date(self, package_info) -> Optional[datetime]:
        """
        Gets the most recent release date of a Python project

        Args:
            releases (dict): PyPI JSON API's representation field

        Returns:
            datetime: creation date of the most recent in releases
        """
        latest_release_version = package_info["dist-tags"]["latest"]
        raw_date = package_info["time"][latest_release_version]
        release_date = parser.isoparse(raw_date).replace(tzinfo=None)
        return release_date
